/* -*- mode: c++; indent-tabs-mode: nil -*- */
/** @file JsonRpcClient.qpp defines the JsonRpcClient class */
/*
  QC_JsonRpcClient.qpp

  Qore Programming Language

  Copyright (C) 2006 - 2012 Qore Technologies

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "qore-json-module.h"

#include "QC_JsonRpcClient.h"

#ifdef _QORE_HAS_QOREHTTPCLIENTOBJECT
#include <qore/QoreHttpClientObject.h>
#define QORE_HTTP_CLIENT QoreHttpClientObject
#else
#include <qore/QoreHTTPClient.h>
#define QORE_HTTP_CLIENT QoreHTTPClient
#endif
#include <qore/ReferenceHolder.h>

#include "ql_json.h"

class HTTPInfoRefHelper {
protected:
   const ReferenceNode *ref;
   ExceptionSink *xsink;
   ReferenceHolder<QoreHashNode> info;

public:
   DLLLOCAL HTTPInfoRefHelper(const ReferenceNode *n_ref, QoreStringNode *msg, ExceptionSink *n_xsink) : ref(n_ref), xsink(n_xsink), info(new QoreHashNode, xsink) {
      info->setKeyValue("request", msg, xsink);
   }
   DLLLOCAL ~HTTPInfoRefHelper() {
      // we have to create a temporary ExceptionSink if there is
      // an active exception, otherwise writing back the reference will fail
      ExceptionSink *txsink = *xsink ? new ExceptionSink : xsink;
      
      // write info hash to reference
      AutoVLock vl(txsink);
      QoreTypeSafeReferenceHelper rh(ref, vl, txsink);
      if (!rh)
         return;

      if (rh.assign(info.release(), txsink))
         return;   

      if (txsink != xsink)
         xsink->assimilate(txsink);
   }
   DLLLOCAL QoreHashNode *operator*() {
      return *info;
   }
};

static void set_jrc_defaults(QORE_HTTP_CLIENT& client) {
   // set encoding to UTF-8
   client.setEncoding(QCS_UTF8);

   // set options for JSON-RPC communication
   client.setDefaultPath("JSON");
   client.setDefaultHeaderValue("Content-Type", "application/json;charset=utf-8");
   client.setDefaultHeaderValue("Accept", "application/json");
   client.setDefaultHeaderValue("User-Agent", "Qore-JSON-RPC-Client/" PACKAGE_VERSION);

   client.addProtocol("jsonrpc", 80, false);
   client.addProtocol("jsonrpcs", 443, true);
}

static AbstractQoreNode *make_jsonrpc_call(QORE_HTTP_CLIENT* client, QoreStringNode *msg, QoreHashNode *info, ExceptionSink *xsink) {
   ReferenceHolder<QoreHashNode> response(client->send("POST", 0, 0, msg->getBuffer(), msg->strlen(), true, info, xsink), xsink);
   if (!response)
      return 0;

   ReferenceHolder<AbstractQoreNode> ans(response->takeKeyValue("body"), xsink);

   if (!ans)
      return 0;

   AbstractQoreNode *ah = *ans;
   if (info) {
      info->setKeyValue("response", ans.release(), xsink);
      info->setKeyValue("response_headers", response.release(), xsink);
   }
   
   if (ah->getType() != NT_STRING) {
      xsink->raiseException("JSONRPCCLIENT-RESPONSE-ERROR", "undecoded binary response received from remote server");
      return 0;
   }

   // parse JSON-RPC response   
   return parseJSONValue(reinterpret_cast<QoreStringNode *>(ah), xsink);
}
/* Qore class Qore::Json::JsonRpcClient */

qore_classid_t CID_JSONRPCCLIENT;
QoreClass* QC_JSONRPCCLIENT;

// hash JsonRpcClient::call(string method, ...){}
static AbstractQoreNode* JsonRpcClient_call_VsVV(QoreObject* self, QORE_HTTP_CLIENT* client, const QoreListNode* args, ExceptionSink* xsink) {
# 234 "QC_JsonRpcClient.qpp"
    // create the outgoing message in JSON-RPC call format
   SimpleRefHolder<QoreStringNode> msg(makeJSONRPC11RequestString(args, xsink));
   if (!msg)
      return 0;

   return make_jsonrpc_call(client, *msg, 0, xsink);
}

// hash JsonRpcClient::callArgs(string method, any args){}
static AbstractQoreNode* JsonRpcClient_callArgs_VsVa(QoreObject* self, QORE_HTTP_CLIENT* client, const QoreListNode* args, ExceptionSink* xsink) {
# 209 "QC_JsonRpcClient.qpp"
    // create the outgoing message in JSON-RPC call format
   SimpleRefHolder<QoreStringNode> msg(makeJSONRPC11RequestStringArgs(args, xsink));
   if (!msg)
      return 0;

   return make_jsonrpc_call(client, *msg, 0, xsink);
}

// hash JsonRpcClient::callArgsWithInfo(reference info, string method, any args){}
static AbstractQoreNode* JsonRpcClient_callArgsWithInfo_VrVsVa(QoreObject* self, QORE_HTTP_CLIENT* client, const QoreListNode* args, ExceptionSink* xsink) {
    const ReferenceNode* info = HARD_QORE_REF(args, 0);
# 268 "QC_JsonRpcClient.qpp"
    // get arguments
   ReferenceHolder<QoreListNode> xargs(args->copyListFrom(1), xsink);

   // create the outgoing message in JSON-RPC call format
   QoreStringNode *msg = makeJSONRPC11RequestStringArgs(*xargs, xsink);
   if (!msg)
      return 0;

   HTTPInfoRefHelper irh(info, msg, xsink);

   // send the message to the server and get the response as a JSON string
   return make_jsonrpc_call(client, msg, *irh, xsink);
}

// hash JsonRpcClient::callWithInfo(reference info, string method, ...){}
static AbstractQoreNode* JsonRpcClient_callWithInfo_VrVsVV(QoreObject* self, QORE_HTTP_CLIENT* client, const QoreListNode* args, ExceptionSink* xsink) {
    const ReferenceNode* info = HARD_QORE_REF(args, 0);
# 307 "QC_JsonRpcClient.qpp"
    // get arguments
   ReferenceHolder<QoreListNode> xargs(args->copyListFrom(1), xsink);

   // create the outgoing message in JSON-RPC call format
   QoreStringNode *msg = makeJSONRPC11RequestString(*xargs, xsink);
   if (!msg)
      return 0;

   HTTPInfoRefHelper irh(info, msg, xsink);

   // send the message to the server and get the response as a JSON string
   return make_jsonrpc_call(client, msg, *irh, xsink);
}

// JsonRpcClient::constructor() {}
static void JsonRpcClient_constructor(QoreObject* self, const QoreListNode* args, ExceptionSink* xsink) {
# 142 "QC_JsonRpcClient.qpp"
// get HTTPClient object
   ReferenceHolder<QORE_HTTP_CLIENT> client((QORE_HTTP_CLIENT*)self->getReferencedPrivateData(CID_HTTPCLIENT, xsink), xsink);
   if (!client)
      return;

   set_jrc_defaults(*(*client));
}

// JsonRpcClient::constructor(hash opts, softbool no_connect = False) {}
static void JsonRpcClient_constructor_Vhvb(QoreObject* self, const QoreListNode* args, ExceptionSink* xsink) {
    const QoreHashNode* opts = HARD_QORE_HASH(args, 0);
    bool no_connect = HARD_QORE_BOOL(args, 1);
# 170 "QC_JsonRpcClient.qpp"
// get HTTPClient object
   ReferenceHolder<QORE_HTTP_CLIENT> client((QORE_HTTP_CLIENT*)self->getReferencedPrivateData(CID_HTTPCLIENT, xsink), xsink);
   if (!client)
      return;

   set_jrc_defaults(*(*client));

   if (client->setOptions(opts, xsink))
      return;

   // do not connect immediately if the second argument is True
   if (!no_connect)
      client->connect(xsink);
}

// JsonRpcClient::copy() {}
static void JsonRpcClient_copy(QoreObject* self, QoreObject* old, QORE_HTTP_CLIENT* client, ExceptionSink* xsink) {
# 189 "QC_JsonRpcClient.qpp"
xsink->raiseException("JSONRPCCLIENT-COPY-ERROR", "copying JsonRpcClient objects is not yet supported.");
}

// nothing JsonRpcClient::setEventQueue(){}
static AbstractQoreNode* JsonRpcClient_setEventQueue(QoreObject* self, QORE_HTTP_CLIENT* client, const QoreListNode* args, ExceptionSink* xsink) {
# 326 "QC_JsonRpcClient.qpp"
    client->setEventQueue(0, xsink);
    return 0;
}

// nothing JsonRpcClient::setEventQueue(Queue queue){}
static AbstractQoreNode* JsonRpcClient_setEventQueue_C5Queue(QoreObject* self, QORE_HTTP_CLIENT* client, const QoreListNode* args, ExceptionSink* xsink) {
    HARD_QORE_OBJ_DATA(queue, Queue, args, 0, CID_QUEUE, "JsonRpcClient::setEventQueue()", "Queue", xsink);
    if (*xsink)
        return 0;
# 337 "QC_JsonRpcClient.qpp"
    // pass reference from QoreObject::getReferencedPrivateData() to function
   client->setEventQueue(queue, xsink);
    return 0;
}

DLLLOCAL void preinitJsonRpcClientClass() {
    QC_JSONRPCCLIENT = new QoreClass("JsonRpcClient", QDOM_DEFAULT);
    CID_JSONRPCCLIENT = QC_JSONRPCCLIENT->getID();
    QC_JSONRPCCLIENT->setSystem();
}

DLLLOCAL QoreClass* initJsonRpcClientClass(QoreNamespace& ns) {
    if (!QC_JSONRPCCLIENT)
        preinitJsonRpcClientClass();

    // set default builtin base class
    assert(QC_HTTPCLIENT);
  QC_JSONRPCCLIENT->addDefaultBuiltinBaseClass(QC_HTTPCLIENT);

    // hash JsonRpcClient::call(string method, ...){}
    QC_JSONRPCCLIENT->addMethodExtended3("call", (q_method_t)JsonRpcClient_call_VsVV, false, QC_USES_EXTRA_ARGS, QDOM_DEFAULT, hashTypeInfo, 1, stringTypeInfo, NULL, "method");

    // hash JsonRpcClient::callArgs(string method, any args){}
    QC_JSONRPCCLIENT->addMethodExtended3("callArgs", (q_method_t)JsonRpcClient_callArgs_VsVa, false, QC_NO_FLAGS, QDOM_DEFAULT, hashTypeInfo, 2, stringTypeInfo, NULL, "method", anyTypeInfo, NULL, "args");

    // hash JsonRpcClient::callArgsWithInfo(reference info, string method, any args){}
    QC_JSONRPCCLIENT->addMethodExtended3("callArgsWithInfo", (q_method_t)JsonRpcClient_callArgsWithInfo_VrVsVa, false, QC_NO_FLAGS, QDOM_DEFAULT, hashTypeInfo, 3, referenceTypeInfo, NULL, "info", stringTypeInfo, NULL, "method", anyTypeInfo, NULL, "args");

    // hash JsonRpcClient::callWithInfo(reference info, string method, ...){}
    QC_JSONRPCCLIENT->addMethodExtended3("callWithInfo", (q_method_t)JsonRpcClient_callWithInfo_VrVsVV, false, QC_USES_EXTRA_ARGS, QDOM_DEFAULT, hashTypeInfo, 2, referenceTypeInfo, NULL, "info", stringTypeInfo, NULL, "method");

    // JsonRpcClient::constructor() {}
    QC_JSONRPCCLIENT->setConstructorExtended3(JsonRpcClient_constructor, false, QC_NO_FLAGS, QDOM_DEFAULT);

    // JsonRpcClient::constructor(hash opts, softbool no_connect = False) {}
    QC_JSONRPCCLIENT->setConstructorExtended3(JsonRpcClient_constructor_Vhvb, false, QC_NO_FLAGS, QDOM_DEFAULT, 2, hashTypeInfo, NULL, "opts", softBoolTypeInfo, &False, "no_connect");

    // JsonRpcClient::copy() {}
    QC_JSONRPCCLIENT->setCopy((q_copy_t)JsonRpcClient_copy);

    // nothing JsonRpcClient::setEventQueue(){}
    QC_JSONRPCCLIENT->addMethodExtended3("setEventQueue", (q_method_t)JsonRpcClient_setEventQueue, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo);

    // nothing JsonRpcClient::setEventQueue(Queue queue){}
    QC_JSONRPCCLIENT->addMethodExtended3("setEventQueue", (q_method_t)JsonRpcClient_setEventQueue_C5Queue, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, QC_QUEUE->getTypeInfo(), NULL, "queue");

    return QC_JSONRPCCLIENT;
}
