/* -*- mode: c++; indent-tabs-mode: nil -*- */
/** @file ql_json.qpp defines JSON functions */
/*
  lib/ql_json.qpp

  Qore JSON (JavaScript QoreObject Notation) functions

  Qore Programming Language

  Copyright 2003 - 2012 David Nichols

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "qore-json-module.h"

#include "ql_json.h"

#include <ctype.h>
#include <stdlib.h>

// RFC 4627 JSON specification
// qore only supports JSON with UTF-8 

// returns 0 for OK
static int cmp_rest_token(const char *&p, const char *tok) {
   p++;
   while (*tok)
      if ((*(p++)) != (*(tok++)))
	 return -1;
   if (!*p || *p == ',' || *p == ']' || *p == '}')
      return 0;
   if (isblank(*p) || (*p) == '\r' || (*p) == '\n') {
      ++p;
      return 0;
   }
   return -1;
} 

static void skip_whitespace(const char *&buf, int &line_number) {
   while (*buf) {
      if (isblank(*buf) || (*buf) == '\r') {
	 ++buf;
	 continue;
      }
      if ((*buf) == '\n') {
	 ++line_number;
	 ++buf;
	 continue;
      }
      break;
   }
}

// '"' has already been read and the buffer is set to this character
static int getJSONStringToken(QoreString &str, const char *&buf, int &line_number, ExceptionSink *xsink) {
   // increment buffer to first character of string
   buf++;
   while (*buf) {
      if (*buf == '"') {
	 buf++;
	 return 0;
      }
      if (*buf == '\\') {
	 buf++;
	 if (*buf == '"' || *buf == '/' || *buf == '\\') {
	    str.concat(*buf);
	    buf++;
	    continue;
	 }
	 if (*buf == 'a')
	    str.concat('\a');
	 else if (*buf == 'b')
	    str.concat('\b');
	 else if (*buf == 'f')
	    str.concat('\f');
	 else if (*buf == 'n')
	    str.concat('\n');
	 else if (*buf == 'r')
	    str.concat('\r');
	 else if (*buf == 't')
	    str.concat('\t');
	 else if (*buf == 'v')
	    str.concat('\v');
	 else if (*buf == 'u') { // expect a unicode character specification
	    ++buf;
	    // check for 4 hex digits
	    if (isxdigit(*buf) && isxdigit(*(buf + 1)) && isxdigit(*(buf + 2)) && isxdigit(*(buf + 3))) {
	       char unicode[5];
	       strncpy(unicode, buf, 4);
	       unicode[4] = '\0';
	       unsigned code = strtoul(unicode, 0, 16);
               if (str.concatUnicode(code, xsink))
                  break;
	       buf += 3;
	    }
	    else
	       str.concat("\\u");
	 }
	 else { // otherwise just concatenate the characters
	    str.concat('\\');
	    str.concat(*buf);
	 }
	 ++buf;
	 continue;
      }
      if (*buf == '\n')
	 line_number++;
      str.concat(*buf);
      ++buf;
   }
   xsink->raiseException("JSON-PARSE-ERROR", "premature end of input at line %d while parsing JSON string", line_number);
   return -1;
}

static AbstractQoreNode *getJSONValue(const char *&buf, int &line_number, const QoreEncoding *enc, ExceptionSink *xsink);

// '{' has already been read and the buffer is set to this character
static QoreHashNode *getJSONObject(const char *&buf, int &line_number, const QoreEncoding *enc, ExceptionSink *xsink) {
   // increment buffer to first character of object description
   buf++;
   ReferenceHolder<QoreHashNode> h(new QoreHashNode(), xsink);

   // get either string or '}'
   skip_whitespace(buf, line_number);
      
   if (*buf == '}') {
      buf++;
      return h.release();
   }

   while (*buf) {
      if (*buf != '"') {
	 //printd(5, "*buf='%c'\n", *buf);
	 if (h->size())
	    xsink->raiseException("JSON-PARSE-ERROR", "unexpected text encountered at line %d while parsing JSON object (expecting '\"' for key string)", line_number);
	 else
	    xsink->raiseException("JSON-PARSE-ERROR", "unexpected text encountered at line %d while parsing JSON object (expecting '\" or '}'')", line_number);
	 break;
      }

      // get key
      QoreString str(enc);
      if (getJSONStringToken(str, buf, line_number, xsink))
	 break;

      //printd(5, "getJSONObject() key=%s\n", str.getBuffer());
      
      skip_whitespace(buf, line_number);
      if (*buf != ':') {
	 //printd(5, "*buf='%c'\n", *buf);
	 xsink->raiseException("JSON-PARSE-ERROR", "unexpected text encountered at line %d while parsing JSON object (expecting ':')", line_number);
	 break;
      }
      buf++;
      skip_whitespace(buf, line_number);

      // get value
      AbstractQoreNode *val = getJSONValue(buf, line_number, enc, xsink);
      if (!val) {
	 if (!xsink->isException())
	    xsink->raiseException("JSON-PARSE-ERROR", "premature end of input at line %d while parsing JSON object (expecting JSON value for key '%s')", line_number, str.getBuffer());
	 break;
      }
      h->setKeyValue(&str, val, xsink);

      skip_whitespace(buf, line_number);
      if (*buf == '}') {
	 buf++;
	 return h.release();
      }

      if (*buf != ',') {
	 xsink->raiseException("JSON-PARSE-ERROR", "unexpected text encountered at line %d while parsing JSON object (expecting ',' or '}')", line_number);
	 break;
      }
      buf++;
      
      skip_whitespace(buf, line_number);

   }
   return 0;
}

// '[' has already been read and the buffer is set to this character
static AbstractQoreNode *getJSONArray(const char *&buf, int &line_number, const QoreEncoding *enc, ExceptionSink *xsink) {
   // increment buffer to first character of array description
   buf++;
   ReferenceHolder<QoreListNode> l(new QoreListNode(), xsink);

   skip_whitespace(buf, line_number);
   if (*buf == ']') {
      ++buf;
      return l.release();
   }

   while (*buf) {
      //printd(5, "before getJSONValue() buf=%s\n", buf);
      AbstractQoreNode *val = getJSONValue(buf, line_number, enc, xsink);
      if (!val) {
	 if (!xsink->isException())
	    xsink->raiseException("JSON-PARSE-ERROR", "premature end of input at line %d while parsing JSON array (expecting JSON value)", line_number);
	 return 0;
      }
      //printd(5, "after getJSONValue() buf=%s\n", buf);
      l->push(val);

      skip_whitespace(buf, line_number);
      if (*buf == ']') {
	 buf++;
	 return l.release();
      }

      if (*buf != ',') {
	 //printd(5, "*buf='%c'\n", *buf);
	 xsink->raiseException("JSON-PARSE-ERROR", "unexpected text encountered at line %d while parsing JSON array (expecting ',' or ']')", line_number);
	 return 0;
      }
      buf++;
      
      skip_whitespace(buf, line_number);
   }
   return 0;
}

static AbstractQoreNode *getJSONValue(const char *&buf, int &line_number, const QoreEncoding *enc, ExceptionSink *xsink) {
   // skip whitespace
   skip_whitespace(buf, line_number);
   if (!*buf)
      return 0;

   // can expect: 't'rue, 'f'alse, '{', '[', '"'string...", integer, '.'digits
   if (*buf == '{')
      return getJSONObject(buf, line_number, enc, xsink);

   if (*buf == '[')
      return getJSONArray(buf, line_number, enc, xsink);

   if (*buf == '"') {
      QoreStringNodeHolder str(new QoreStringNode(enc));
      return getJSONStringToken(*(*str), buf, line_number, xsink) ? 0 : str.release();
   }

   // FIXME: implement parsing of JSON exponents
   if (isdigit(*buf) || (*buf) == '.' || (*buf) == '-') {
      // temporarily use a QoreString
      QoreString str;
      bool has_dot;
      if (*buf == '.') {
	 // add a leading zero
	 str.concat("0.");
	 has_dot = true;
      }
      else {
	 str.concat(*buf);
	 has_dot = false;
      }
      buf++;
      while (*buf) {
	 if (*buf == '.') {
	    if (has_dot) {
	       xsink->raiseException("JSON-PARSE-ERROR", "unexpected '.' in floating point number (too many '.' characters)");
	       return 0;
	    }
	    has_dot = true;
	 }
	 // if another token follows then break but do not increment buffer position
	 else if (*buf == ',' || *buf == '}' || *buf == ']')
	    break;
	 // if whitespace follows then increment buffer position and break
	 else if (isblank(*buf) || (*buf) == '\r') {
	    ++buf;
	    break;
	 }
	 // if a newline follows then  increment buffer position and line number and break
	 else if ((*buf) == '\n') {
	    ++buf;
	    ++line_number;
	    break;
	 }
	 else if (!isdigit(*buf)) {
	    xsink->raiseException("JSON-PARSE-ERROR", "unexpected character in number");
	    return 0;
	 }
	 str.concat(*buf);
	 buf++;
      }
      if (has_dot)
	 return new QoreFloatNode(strtod(str.getBuffer(), 0));
      return new QoreBigIntNode(strtoll(str.getBuffer(), 0, 10));
   }
   
   if ((*buf) == 't') {
      if (!cmp_rest_token(buf, "rue"))
	 return boolean_true();
      goto error;
   }
   if ((*buf) == 'f') {
      if (!cmp_rest_token(buf, "alse"))
	 return boolean_false();
      goto error;
   }
   if ((*buf) == 'n') {
      if (!cmp_rest_token(buf, "ull"))
	 return nothing();
      goto error;
   }
   
  error:
   //printd(5, "buf=%s\n", buf);

   xsink->raiseException("JSON-PARSE-ERROR", "invalid input at line %d; unable to parse JSON value", line_number);
   return 0;
}

#define JSF_THRESHOLD 20

static int doJSONValue(QoreString *str, const AbstractQoreNode *v, int format, ExceptionSink *xsink) {
   if (is_nothing(v) || is_null(v)) {
      str->concat("null");
      return 0;
   }

   qore_type_t vtype = v->getType();

   if (vtype == NT_LIST) {
      const QoreListNode *l = reinterpret_cast<const QoreListNode *>(v);
      str->concat("[ ");
      ConstListIterator li(l);
      QoreString tmp(str->getEncoding());
      while (li.next()) {
	 bool ind = tmp.strlen() > JSF_THRESHOLD;
	 tmp.clear();
	 if (doJSONValue(&tmp, li.getValue(), format == -1 ? format : format + 2, xsink))
	    return -1;
	 
	 if (format != -1 && (ind || tmp.strlen() > JSF_THRESHOLD)) {
	    str->concat('\n');
	    str->addch(' ', format + 2);
	 }
	 str->sprintf("%s", tmp.getBuffer());
	 
	 if (!li.last())
	    str->concat(", ");
      }
      str->concat(" ]");
      return 0;
   }

   if (vtype == NT_HASH) {
      const QoreHashNode *h = reinterpret_cast<const QoreHashNode *>(v);
      str->concat("{ ");
      ConstHashIterator hi(h);
      QoreString tmp(str->getEncoding());
      while (hi.next()) {
	 bool ind = tmp.strlen() > JSF_THRESHOLD;
	 tmp.clear();
	 if (doJSONValue(&tmp, hi.getValue(), format == -1 ? format : format + 2, xsink))
	    return -1;
	 
	 if (format != -1 && (ind || tmp.strlen() > JSF_THRESHOLD)) {
	    str->concat('\n');
	    str->addch(' ', format + 2);
	 }
	 str->sprintf("\"%s\" : %s", hi.getKey(), tmp.getBuffer());
	 if (!hi.last())
	    str->concat(", ");
      }
      str->concat(" }");
      return 0;
   }

   if (vtype == NT_STRING) {
      const QoreStringNode *vstr = reinterpret_cast<const QoreStringNode*>(v);

      TempEncodingHelper t(vstr, str->getEncoding(), xsink);
      if (*xsink)
	 return -1;
      
      str->concat('"');
      qore_size_t i = str->size();
      str->concatEscape(*t, '"', '\\', xsink);
      if (*xsink)
	 return -1;

      // http://tools.ietf.org/html/rfc4627
      // encode all control characters in the string concatenated
      while (i < str->size()) {
         // see if we have a single-byte character; do not try to look at bytes within multi-byte characters
         qore_size_t cl = q_get_char_len(str->getEncoding(), str->getBuffer() + i, str->size() - i, xsink);
         if (*xsink)
            return -1;
         if (cl > 1) {
            i += cl;
            continue;
         }

         unsigned char c = (*str)[i];         
         if (c < 32) {
            switch (c) {
               case 7: str->replace(i, 1, "\\a"); ++i; break; // BEL
               case 8: str->replace(i, 1, "\\b"); ++i; break; // BS
               case 9: str->replace(i, 1, "\\t"); ++i; break; // HT
               case 10: str->replace(i, 1, "\\n"); ++i; break; // LF
               case 11: str->replace(i, 1, "\\v"); ++i; break; // VT
               case 12: str->replace(i, 1, "\\f"); ++i; break; // FF
               case 13: str->replace(i, 1, "\\r"); ++i; break; // CR
               default: {
                  char buf[7];
                  sprintf(buf, "\\u%04x", (int)c);
                  str->replace(i, 1, buf); i += 5;
                  break;
               }
            }
         }
         ++i;
      }

      str->concat('"');

      return 0;
   }

   if (vtype == NT_INT) {
      const QoreBigIntNode *b = reinterpret_cast<const QoreBigIntNode *>(v);
      str->sprintf("%lld", b->val);
      return 0;
   }

   if (vtype == NT_FLOAT) {
      str->sprintf("%.20g", reinterpret_cast<const QoreFloatNode *>(v)->f);
      return 0;
   }

#ifdef _QORE_HAS_NUMBER_TYPE
   if (vtype == NT_NUMBER) {
      reinterpret_cast<const QoreNumberNode*>(v)->getStringRepresentation(*str);
      return 0;
   }
#endif

   if (vtype == NT_BOOLEAN) {
      str->concat(reinterpret_cast<const QoreBoolNode*>(v)->getValue() ? "true" : "false");
      return 0;
   }

   if (vtype == NT_DATE) {
      const DateTimeNode *date = reinterpret_cast<const DateTimeNode*>(v);
      // this will be serialized as a string
      str->concat('"');
#ifdef _QORE_HAS_TIME_ZONES

#ifndef SECS_PER_MINUTE
#define SECS_PER_MINUTE          60
#endif
#ifndef SECS_PER_HOUR
#define SECS_PER_HOUR            (SECS_PER_MINUTE * 60)
#endif

      // ensure that all date/time values are reported in the current time zone
      // for simplicity's sake (particularly because json does not have a native date/time format)
      qore_tm info;
      date->getInfo(currentTZ(), info);
      str->sprintf("%04d-%02d-%02d %02d:%02d:%02d.%06d", info.year, info.month, info.day, info.hour, info.minute, info.second, info.us);
      if (!info.utc_secs_east)
         str->concat('Z');
      else {
         str->concat(' ');
         str->concat(info.utc_secs_east < 0 ? '-' : '+');
         if (info.utc_secs_east < 0)
            info.utc_secs_east = -info.utc_secs_east;
         int h = info.utc_secs_east / SECS_PER_HOUR;
         // the remaining seconds after hours
         int r = info.utc_secs_east % SECS_PER_HOUR;
         // minutes
         int m = r / SECS_PER_MINUTE;
         // we have already output the hour sign above
         str->sprintf("%02d:%02d", h < 0 ? -h : h, m);
         // see if there are any seconds
         int s = info.utc_secs_east - h * SECS_PER_HOUR - m * SECS_PER_MINUTE;
         if (s)
            str->sprintf(":%02d", s);
      }
#else
      date->DateTimeNode::getAsString(*str, 0, 0);
#endif
      str->concat('"');
      return 0;
   }
   
   xsink->raiseException("JSON-SERIALIZATION-ERROR", "don't know how to serialize type '%s'", v->getTypeName());
   return -1;
}

QoreStringNode *makeJSONRPC11RequestStringArgs(const QoreListNode *params, ExceptionSink *xsink) {
   const QoreStringNode *p0;
   if (!(p0 = test_string_param(params, 0))) {
      xsink->raiseException("MAKE-JSONRPC11-REQUEST-STRING-ERROR", "expecting method name as first parameter");
      return 0;
   }

   const AbstractQoreNode *p1 = get_param(params, 1);

   QoreStringNodeHolder str(new QoreStringNode(QCS_UTF8));

   // write version key first
   str->concat("{ \"version\" : \"1.1\", \"method\" : ");
   if (doJSONValue(*str, p0, -1, xsink))
      return 0;

   // params key should come last
   str->concat(", \"params\" : ");
   if (!is_nothing(p1)) {
      if (doJSONValue(*str, p1, -1, xsink))
	 return 0;
   }
   else
      str->concat("null");
   str->concat(" }");
   return str.release();
}

QoreStringNode *makeJSONRPC11RequestString(const QoreListNode *params, ExceptionSink *xsink) {
   const QoreStringNode *p0;
   if (!(p0 = test_string_param(params, 0))) {
      xsink->raiseException("MAKE-JSONRPC11-REQUEST-STRING-ERROR", "expecting method name as first parameter");
      return 0;
   }

   QoreStringNodeHolder str(new QoreStringNode(QCS_UTF8));

   // write version key first
   str->concat("{ \"version\" : \"1.1\", \"method\" : ");

   if (doJSONValue(*str, p0, -1, xsink))
      return 0;

   // params key should come last
   str->concat(", \"params\" : ");
   if (num_params(params) > 1) {
      ReferenceHolder<QoreListNode> new_params(params->copyListFrom(1), xsink);

      if (doJSONValue(*str, *new_params, -1, xsink))
	 return 0;
   }
   else
      str->concat("null");
   str->concat(" }");

   return str.release();
}

AbstractQoreNode *parseJSONValue(const QoreString *str, ExceptionSink *xsink) {
   int line_number = 1;
   const char *buf = str->getBuffer();
   AbstractQoreNode *rv = getJSONValue(buf, line_number, str->getEncoding(), xsink);
   if (rv && *buf) {
      // check for excess text after JSON data
      skip_whitespace(buf, line_number);
      if (*buf) {
	 xsink->raiseException("JSON-PARSE-ERROR", "extra text after JSON data on line %d", line_number);
	 rv->deref(xsink);
	 rv = 0;
      }
   }
   return rv;
}
# 586 "ql_json.qpp"
# 591 "ql_json.qpp"
# 611 "ql_json.qpp"
# 631 "ql_json.qpp"
# 637 "ql_json.qpp"
// string makeFormattedJSONString(any data, *string encoding) {}
static AbstractQoreNode* f_makeFormattedJSONString_VaNs(const QoreListNode* args, ExceptionSink* xsink) {
    const AbstractQoreNode* data = get_param(args, 0);
    const QoreStringNode* encoding = reinterpret_cast<const QoreStringNode*>(get_param(args, 1));
# 625 "ql_json.qpp"
   const QoreEncoding *qe = encoding ? QEM.findCreate(encoding) : QCS_UTF8;
   QoreStringNodeHolder str(new QoreStringNode(qe));
   return doJSONValue(*str, data, 0, xsink) ? 0 : str.release();
}

// string makeJSONString(any data, *string encoding) {}
static AbstractQoreNode* f_makeJSONString_VaNs(const QoreListNode* args, ExceptionSink* xsink) {
    const AbstractQoreNode* data = get_param(args, 0);
    const QoreStringNode* encoding = reinterpret_cast<const QoreStringNode*>(get_param(args, 1));
# 605 "ql_json.qpp"
   const QoreEncoding *qe = encoding ? QEM.findCreate(encoding) : QCS_UTF8;
   QoreStringNodeHolder str(new QoreStringNode(qe));
   return doJSONValue(*str, data, -1, xsink) ? 0 : str.release();
}

// nothing parseJSON() {}
static AbstractQoreNode* f_parseJSON(const QoreListNode* args, ExceptionSink* xsink) {
# 634 "ql_json.qpp"

    return 0;
}

// any parseJSON(string json_str) {}
static AbstractQoreNode* f_parseJSON_Vs(const QoreListNode* args, ExceptionSink* xsink) {
    const QoreStringNode* json_str = HARD_QORE_STRING(args, 0);
# 649 "ql_json.qpp"
   return parseJSONValue(json_str, xsink);
}

# 654 "ql_json.qpp"
# 707 "ql_json.qpp"
# 757 "ql_json.qpp"
# 802 "ql_json.qpp"
# 846 "ql_json.qpp"
# 891 "ql_json.qpp"
# 936 "ql_json.qpp"
# 995 "ql_json.qpp"
// string makeFormattedJSONRPC11ErrorString(softint code, string error, any id, any error_msg) {}
static AbstractQoreNode* f_makeFormattedJSONRPC11ErrorString_viVsVaVa(const QoreListNode* args, ExceptionSink* xsink) {
    int64 code = HARD_QORE_INT(args, 0);
    const QoreStringNode* error = HARD_QORE_STRING(args, 1);
    const AbstractQoreNode* id = get_param(args, 2);
    const AbstractQoreNode* error_msg = get_param(args, 3);
# 1013 "ql_json.qpp"
   if (code < 100 || code > 999) {
      xsink->raiseException("MAKE-JSONRPC11-ERROR-STRING-ERROR", "error code (first argument) must be between 100 and 999 inclusive (value passed: "QLLD")", code);
      return 0;
   }
   
   if (error->empty()) {
      xsink->raiseException("MAKE-JSONRPC11-ERROR-STRING-ERROR", "error message string not passed as second argument)");
      return 0;
   }

   QoreStringNodeHolder str(new QoreStringNode(QCS_UTF8));
   str->sprintf("{\n  \"version\" : \"1.1\",\n  ");

   // get optional "id" value
   if (!is_nothing(id)) {
      str->concat("\"id\" : ");
      if (doJSONValue(*str, id, -1, xsink))
	 return 0;
      str->concat(",\n  ");
   }

   str->sprintf("\"error\" :\n  {\n    \"name\" : \"JSONRPCError\",\n    \"code\" : %d,\n    \"message\" : \"", (int)code);
   // concat here so character encodings can be automatically converted if necessary
   str->concatEscape(error, '"', '\\', xsink);
   if (*xsink)
      return 0;

   str->concat('\"');

   // get optional "error" value
   if (!is_nothing(error_msg)) {
      str->concat(",\n    \"error\" : ");
      if (doJSONValue(*str, error_msg, 4, xsink))
	 return 0;
   }
   str->concat("\n  }\n}");
   return str.release();
}

// string makeFormattedJSONRPCErrorString(any version, any id, any error_msg) {}
static AbstractQoreNode* f_makeFormattedJSONRPCErrorString_VaVaVa(const QoreListNode* args, ExceptionSink* xsink) {
    const AbstractQoreNode* version = get_param(args, 0);
    const AbstractQoreNode* id = get_param(args, 1);
    const AbstractQoreNode* error_msg = get_param(args, 2);
# 907 "ql_json.qpp"
   QoreStringNodeHolder str(new QoreStringNode(QCS_UTF8));

   // write version key first if present
   if (!is_nothing(version)) {
      str->concat("{\n  \"version\" : ");
      if (doJSONValue(*str, version, 2, xsink))
	 return 0;
      str->concat(",\n  ");
   }
   else
      str->concat("{\n  ");

   if (!is_nothing(id)) {
      str->concat("\"id\" : ");
      if (doJSONValue(*str, id, 2, xsink))
	 return 0;
      str->concat(",\n  ");
   }

   // error key should come last
   str->concat("\"error\" : ");
   if (doJSONValue(*str, error_msg, 2, xsink))
      return 0;

   str->concat("\n}");
   return str.release();
}

// string makeFormattedJSONRPCRequestString(string method_name, any version, any id, any request_msg) {}
static AbstractQoreNode* f_makeFormattedJSONRPCRequestString_VsVaVaVa(const QoreListNode* args, ExceptionSink* xsink) {
    const QoreStringNode* method_name = HARD_QORE_STRING(args, 0);
    const AbstractQoreNode* version = get_param(args, 1);
    const AbstractQoreNode* id = get_param(args, 2);
    const AbstractQoreNode* request_msg = get_param(args, 3);
# 725 "ql_json.qpp"
   QoreStringNodeHolder str(new QoreStringNode(QCS_UTF8));

   // write version key first if present
   if (!is_nothing(version)) {
      str->concat("{\n  \"version\" : ");
      if (doJSONValue(*str, version, 2, xsink))
	 return 0;
      str->concat(",\n  ");
   }
   else
      str->concat("{\n  ");

   str->concat("\"method\" : ");
   if (doJSONValue(*str, method_name, 2, xsink))
      return 0;

   if (!is_nothing(id)) {
      str->concat(",\n  \"id\" : ");
      if (doJSONValue(*str, id, 2, xsink))
	 return 0;
   }

   // params key should come last
   str->concat(",\n  \"params\" : ");
   if (doJSONValue(*str, request_msg, 2, xsink))
      return 0;

   str->concat("\n}");
   return str.release();
}

// string makeFormattedJSONRPCResponseString(any version, any id, any response_msg) {}
static AbstractQoreNode* f_makeFormattedJSONRPCResponseString_VaVaVa(const QoreListNode* args, ExceptionSink* xsink) {
    const AbstractQoreNode* version = get_param(args, 0);
    const AbstractQoreNode* id = get_param(args, 1);
    const AbstractQoreNode* response_msg = get_param(args, 2);
# 818 "ql_json.qpp"
   QoreStringNodeHolder str(new QoreStringNode(QCS_UTF8));

   // write version key first if present
   if (!is_nothing(version)) {
      str->concat("{\n  \"version\" : ");
      if (doJSONValue(*str, version, 2, xsink))
	 return 0;
      str->concat(",\n  ");
   }
   else
      str->concat("{\n  ");

   if (!is_nothing(id)) {
      str->concat("\"id\" : ");
      if (doJSONValue(*str, id, 2, xsink))
	 return 0;
      str->concat(",\n  ");
   }

   // result key should come last
   str->concat("\"result\" : ");
   if (doJSONValue(*str, response_msg, 2, xsink))
      return 0;
   str->concat("\n}");
   return str.release();
}

// string makeJSONRPC11ErrorString(softint code, string error, any id, any error_msg) {}
static AbstractQoreNode* f_makeJSONRPC11ErrorString_viVsVaVa(const QoreListNode* args, ExceptionSink* xsink) {
    int64 code = HARD_QORE_INT(args, 0);
    const QoreStringNode* error = HARD_QORE_STRING(args, 1);
    const AbstractQoreNode* id = get_param(args, 2);
    const AbstractQoreNode* error_msg = get_param(args, 3);
# 954 "ql_json.qpp"
   if (code < 100 || code > 999) {
      xsink->raiseException("MAKE-JSONRPC11-ERROR-STRING-ERROR", "error code (first argument) must be between 100 and 999 inclusive (value passed: "QLLD")", code);
      return 0;
   }

   if (error->empty()) {
      xsink->raiseException("MAKE-JSONRPC11-ERROR-STRING-ERROR", "empty error message string passed as second argument)");
      return 0;
   }

   QoreStringNodeHolder str(new QoreStringNode(QCS_UTF8));

   str->concat("{ \"version\" : \"1.1\", ");

   // get optional "id" value
   if (!is_nothing(id)) {
      str->concat("\"id\" : ");
      if (doJSONValue(*str, id, -1, xsink))
	 return 0;
      str->concat(", ");
   }
   
   str->sprintf("\"error\" : { \"name\" : \"JSONRPCError\", \"code\" : %d, \"message\" : \"", (int)code);
   // concat here so character encodings can be automatically converted if necessary
   str->concatEscape(error, '"', '\\', xsink);
   if (xsink->isException())
      return 0;

   str->concat('\"');

   // get optional "error" value
   if (!is_nothing(error_msg)) {
      str->concat(", \"error\" : ");
      if (doJSONValue(*str, error_msg, -1, xsink))
	 return 0;
   }
   str->concat(" } }");
   return str.release();
}

// string makeJSONRPCErrorString(any version, any id, any error_msg) {}
static AbstractQoreNode* f_makeJSONRPCErrorString_VaVaVa(const QoreListNode* args, ExceptionSink* xsink) {
    const AbstractQoreNode* version = get_param(args, 0);
    const AbstractQoreNode* id = get_param(args, 1);
    const AbstractQoreNode* error_msg = get_param(args, 2);
# 862 "ql_json.qpp"
   QoreStringNodeHolder str(new QoreStringNode(QCS_UTF8));

   // write version key first if present
   if (!is_nothing(version)) {
      str->concat("{ \"version\" : ");
      if (doJSONValue(*str, version, -1, xsink))
	 return 0;
      str->concat(", ");
   }
   else
      str->concat("{ ");

   if (!is_nothing(id)) {
      str->concat("\"id\" : ");
      if (doJSONValue(*str, id, -1, xsink))
	 return 0;
      str->concat(", ");
   }

   // error key should come last
   str->concat("\"error\" : ");
   if (doJSONValue(*str, error_msg, -1, xsink))
      return 0;

   str->concat(" }");
   return str.release();
}

// string makeJSONRPCRequestString(string method_name, any version, any id, any request_msg) {}
static AbstractQoreNode* f_makeJSONRPCRequestString_VsVaVaVa(const QoreListNode* args, ExceptionSink* xsink) {
    const QoreStringNode* method_name = HARD_QORE_STRING(args, 0);
    const AbstractQoreNode* version = get_param(args, 1);
    const AbstractQoreNode* id = get_param(args, 2);
    const AbstractQoreNode* request_msg = get_param(args, 3);
# 675 "ql_json.qpp"
   QoreStringNodeHolder str(new QoreStringNode(QCS_UTF8));

   // write version key first if present
   if (!is_nothing(version)) {
      str->concat("{ \"version\" : ");
      if (doJSONValue(*str, version, -1, xsink))
	 return 0;
      str->concat(", ");
   }
   else
      str->concat("{ ");

   str->concat("\"method\" : ");
   if (doJSONValue(*str, method_name, -1, xsink))
      return 0;

   if (!is_nothing(id)) {
      str->concat(", \"id\" : ");
      if (doJSONValue(*str, id, -1, xsink))
	 return 0;
   }

   // params key should come last
   str->concat(",  \"params\" : ");
   if (doJSONValue(*str, request_msg, -1, xsink))
      return 0;

   str->concat(" }");
   return str.release();
}

// string makeJSONRPCResponseString(any version, any id, any response_msg) {}
static AbstractQoreNode* f_makeJSONRPCResponseString_VaVaVa(const QoreListNode* args, ExceptionSink* xsink) {
    const AbstractQoreNode* version = get_param(args, 0);
    const AbstractQoreNode* id = get_param(args, 1);
    const AbstractQoreNode* response_msg = get_param(args, 2);
# 773 "ql_json.qpp"
   QoreStringNodeHolder str(new QoreStringNode(QCS_UTF8));

   // write version key first if present
   if (!is_nothing(version)) {
      str->concat("{ \"version\" : ");
      if (doJSONValue(*str, version, -1, xsink))
	 return 0;
      str->concat(", ");
   }
   else
      str->concat("{ ");

   if (!is_nothing(id)) {
      str->concat("\"id\" : ");
      if (doJSONValue(*str, id, -1, xsink))
	 return 0;
      str->concat(", ");
   }

   // result key should come last
   str->concat("\"result\" : ");
   if (doJSONValue(*str, response_msg, -1, xsink))
      return 0;

   str->concat(" }");
   return str.release();
}


DLLLOCAL void init_json_functions(QoreNamespace& ns) {
    // string makeFormattedJSONString(any data, *string encoding) {}
    ns.addBuiltinVariant("makeFormattedJSONString", (q_func_t)f_makeFormattedJSONString_VaNs, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 2, anyTypeInfo, NULL, "data", stringOrNothingTypeInfo, NULL, "encoding");
    // string makeJSONString(any data, *string encoding) {}
    ns.addBuiltinVariant("makeJSONString", (q_func_t)f_makeJSONString_VaNs, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 2, anyTypeInfo, NULL, "data", stringOrNothingTypeInfo, NULL, "encoding");
    // nothing parseJSON() {}
    ns.addBuiltinVariant("parseJSON", (q_func_t)f_parseJSON, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
    // any parseJSON(string json_str) {}
    ns.addBuiltinVariant("parseJSON", (q_func_t)f_parseJSON_Vs, QC_RET_VALUE_ONLY, QDOM_DEFAULT, anyTypeInfo, 1, stringTypeInfo, NULL, "json_str");
    // string makeFormattedJSONRPC11ErrorString(softint code, string error, any id, any error_msg) {}
    ns.addBuiltinVariant("makeFormattedJSONRPC11ErrorString", (q_func_t)f_makeFormattedJSONRPC11ErrorString_viVsVaVa, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 4, softBigIntTypeInfo, NULL, "code", stringTypeInfo, NULL, "error", anyTypeInfo, NULL, "id", anyTypeInfo, NULL, "error_msg");
    // string makeFormattedJSONRPCErrorString(any version, any id, any error_msg) {}
    ns.addBuiltinVariant("makeFormattedJSONRPCErrorString", (q_func_t)f_makeFormattedJSONRPCErrorString_VaVaVa, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 3, anyTypeInfo, NULL, "version", anyTypeInfo, NULL, "id", anyTypeInfo, NULL, "error_msg");
    // string makeFormattedJSONRPCRequestString(string method_name, any version, any id, any request_msg) {}
    ns.addBuiltinVariant("makeFormattedJSONRPCRequestString", (q_func_t)f_makeFormattedJSONRPCRequestString_VsVaVaVa, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 4, stringTypeInfo, NULL, "method_name", anyTypeInfo, NULL, "version", anyTypeInfo, NULL, "id", anyTypeInfo, NULL, "request_msg");
    // string makeFormattedJSONRPCResponseString(any version, any id, any response_msg) {}
    ns.addBuiltinVariant("makeFormattedJSONRPCResponseString", (q_func_t)f_makeFormattedJSONRPCResponseString_VaVaVa, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 3, anyTypeInfo, NULL, "version", anyTypeInfo, NULL, "id", anyTypeInfo, NULL, "response_msg");
    // string makeJSONRPC11ErrorString(softint code, string error, any id, any error_msg) {}
    ns.addBuiltinVariant("makeJSONRPC11ErrorString", (q_func_t)f_makeJSONRPC11ErrorString_viVsVaVa, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 4, softBigIntTypeInfo, NULL, "code", stringTypeInfo, NULL, "error", anyTypeInfo, NULL, "id", anyTypeInfo, NULL, "error_msg");
    // string makeJSONRPCErrorString(any version, any id, any error_msg) {}
    ns.addBuiltinVariant("makeJSONRPCErrorString", (q_func_t)f_makeJSONRPCErrorString_VaVaVa, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 3, anyTypeInfo, NULL, "version", anyTypeInfo, NULL, "id", anyTypeInfo, NULL, "error_msg");
    // string makeJSONRPCRequestString(string method_name, any version, any id, any request_msg) {}
    ns.addBuiltinVariant("makeJSONRPCRequestString", (q_func_t)f_makeJSONRPCRequestString_VsVaVaVa, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 4, stringTypeInfo, NULL, "method_name", anyTypeInfo, NULL, "version", anyTypeInfo, NULL, "id", anyTypeInfo, NULL, "request_msg");
    // string makeJSONRPCResponseString(any version, any id, any response_msg) {}
    ns.addBuiltinVariant("makeJSONRPCResponseString", (q_func_t)f_makeJSONRPCResponseString_VaVaVa, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 3, anyTypeInfo, NULL, "version", anyTypeInfo, NULL, "id", anyTypeInfo, NULL, "response_msg");

}
